/****************************************************************************
*	$Id: fingerd.c 1.3 1996/10/22 21:00:56 Graham Exp $
*	01 Feb 93	1.2		GT	New sendfile () prototype.						*
*	   Oct 96	1.3		GeK	Improved logging, recognise /W (finger -l),		*
*	         	   		   	added recv fail check							*
****************************************************************************/

/* Internet Finger server
 * Copyright 1991 Phil Karn, KA9Q
 */
#include <stdio.h>
#include <string.h>
#include "global.h"
#include "files.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "proc.h"
#include "dirutil.h"
#include "commands.h"
#include "mailbox.h"

static int Sfinger = -1;	/* Prototype socket for service */

static void fingerd __ARGS((int s,void *unused,void *p));

/* Start up finger service */
int
finstart(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct sockaddr_in lsocket;
	int s;

	if(Sfinger != -1){
		return 0;
	}
	psignal(Curproc,0);	/* Don't keep the parser waiting */
	chname(Curproc,"Finger listener");

	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	if(argc < 2)
		lsocket.sin_port = IPPORT_FINGER;
	else
		lsocket.sin_port = atoi(argv[1]);

	Sfinger = socket(AF_INET,SOCK_STREAM,0);
	bind(Sfinger,(char *)&lsocket,sizeof(lsocket));
	listen(Sfinger,1);
	for(;;){
		if((s = accept(Sfinger,NULLCHAR,(int *)NULL)) == -1)
			break;	/* Service is shutting down */

		/* Spawn a server */
		newproc("fingerd",512,fingerd,s,NULL,NULL,0);
	}
	return 0;
}
static void
fingerd(s,unused,p)
int s;
void *unused;
void *p;
{
	char buf[80];		/*  GeK - buffer  */
	char *user = buf;	/*  GeK - pointer to track username  */
	FILE *fp;
	char *file,*cp;

	sockmode(s,SOCK_ASCII);
	sockowner(s,Curproc);
	log(s,"open Finger");
	if (recvline(s,buf,sizeof(buf)) == -1)		/*  GeK - check for failure  */
		goto quit;
	rip(buf);
	log(s, "fingerd: %s", buf);		/*  GeK - log who's asking what  */
	/*  GeK - skip -l  */
	if (buf[0] == '/' && (buf[1] == 'W' || buf[1] == 'w')) {
		user += 2;
		while (*user == ' ' || *user == '\t')
			user++;
	}
	if(*user == '\0'){
		fp = dir(Fdir,0);
		if(fp == NULLFILE)
			usputs(s,"No finger information available\n");
		else
			usputs(s,"Known users on this system:\n");
	} else {
		file = pathname(Fdir,user);
		cp = pathname(Fdir,"");
		/* Check for attempted security violation (e.g., somebody
		 * might be trying to finger "../ftpusers"!)
		 */
		if(strncmp(file,cp,strlen(cp)) != 0){
			fp = NULLFILE;
			usprintf(s,"Invalid user name %s\n",user);
		} else if((fp = fopen(file,READ_TEXT)) == NULLFILE)
			usprintf(s,"User %s not known\n",user);
		free(cp);
		free(file);
	}
	if(fp != NULLFILE){
		sendfile(fp,s,ASCII_TYPE,0, 0);
		fclose(fp);
	}
	if(*user == '\0' && Listusers != NULL)
		(*Listusers)(s);
quit:
	log(s,"close Finger");		/*  GeK - moved above close_s(s)  */
	close_s(s);
}
int
fin0(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	close_s(Sfinger);
	Sfinger = -1;
	return 0;
}
