/* Machine or compiler-dependent portions of kernel
 * Turbo-C version for PC
 *
 * Copyright 1991 Phil Karn, KA9Q
 *
 * $Id: ksubr.c 1.4 1997/11/24 22:39:12 Graham Exp $
 *
 *	30 Jan 93	1.2		GT	Make stkutil () public when doing PROCTRACE or
 *							PROCLOG.
 *	   May 97	1.3		GeK	Removed env struct & replaced with macros to
 *	         	   		   	access jmpbuf, added OS/2 tasker support,
 *	         	   		   	Tidied tprintfs, add PSP to ps output,
 *	         	   		   	fixed psetup() prototype, added STACK_CHECK
 *	   Nov 97	1.4		GeK	Clear BP in psetup() for stack traceing
 */

#include <stdio.h>
#include <dos.h>
#include <malloc.h>
#include "global.h"
#include "proc.h"
#include "pc.h"
#include "commands.h"
#include "socket.h"		/*  GeK - tputc, tputs  */

static char *Taskers[] = {
	"",
	"DoubleDos",
	"DesqView",
	"Windows",
	"OS/2",
};


static unsigned short	oldNull;	/*  GeK - added missing type  */

#ifdef	notdef	/*  GeK - remove this since it doesn't match BC4  */
/* Template for contents of jmp_buf in Turbo C */
struct env {
	unsigned	sp;
	unsigned	ss;
	unsigned	flag;
	unsigned	cs;
	unsigned	ip;
	unsigned	bp;
	unsigned	di;
	unsigned	es;
	unsigned	si;
	unsigned	ds;
};
#endif	/* notdef */

/*  GeK - replacement macros to access jmpbuf contents with a little portability  */
#define JMPBUF_SS(jbp)	((jbp)->j_ss)
#define JMPBUF_SP(jbp)	((jbp)->j_sp)
#define JMPBUF_CS(jbp)	((jbp)->j_cs)
#define JMPBUF_IP(jbp)	((jbp)->j_ip)
#define JMPBUF_BP(jbp)	((jbp)->j_bp)

/*  GeK - macro to return a far ptr for the stack ptr  */
#define JMPBUF_FP_SSSP(jbp)	MK_FP (JMPBUF_SS (jbp), JMPBUF_SP (jbp))

static int chkintstk __ARGS((void));
#if	defined (PROCTRACE) || defined (PROCLOG)
int stkutil __ARGS((struct proc *pp));
#else
static int stkutil __ARGS((struct proc *pp));
#endif

void
kinit()
{
	int i;

	/* Initialize interrupt stack for high-water-mark checking */
	for(i=0;i<512;i++)
		Intstk[i] = STACKPAT;

	/* Remember location 0 pattern to detect null pointer derefs */
	oldNull = *(unsigned short *)NULL;

}
/* Print process table info
 * Since things can change while ps is running, the ready proceses are
 * displayed last. This is because an interrupt can make a process ready,
 * but a ready process won't spontaneously become unready. Therefore a
 * process that changes during ps may show up twice, but this is better
 * than not having it showing up at all.
 */
int
ps(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	register struct proc *pp;
#ifdef	notdef	/*  GeK - removed env  */
	register struct env *ep;
#endif	/* notdef */
	int i;

	tprintf("Uptime %s Stack %x max intstk %u psp %04x",tformat(secclock()),
	 getss(),chkintstk(),_psp);	/*  GeK - added psp from PK's export v951123  */
	if(Mtasker != 0){
		tprintf(" Running under %s",Taskers[Mtasker]);
	}
	tputc('\n');

	tputs("PID       SP        stksize   maxstk    event     fl  in  out  name\n");

	for(pp = Susptab;pp != NULLPROC;pp = pp->next){
#ifdef	notdef	/*  GeK - removed env  */
		ep = (struct env *)&pp->env;
#endif	/* notdef */
		if(tprintf("%-10lx%-10lx%-10u%-10u%-10lx%c%c%c %3d %3d  %s\n",
		 ptol(pp),
		 ptol(JMPBUF_FP_SSSP (pp->env)),	/*  GeK - JMPBUF macro for portability  */
		 pp->stksize,
		 stkutil(pp),
		 ptol(pp->event),
		 pp->i_state ? 'I' : ' ',
		 (pp->state & WAITING) ? 'W' : ' ',
		 (pp->state & SUSPEND) ? 'S' : ' ',
		 pp->input, pp->output,
		 pp->name) == EOF)
			return 0;
	}
	for(i=0;i<PHASH;i++){
		for(pp = Waittab[i];pp != NULLPROC;pp = pp->next){
#ifdef	notdef	/*  GeK - removed env  */
			ep = (struct env *)&pp->env;
#endif	/* notdef */
			if(tprintf("%-10lx%-10lx%-10u%-10u%-10lx%c%c%c %2d %2d  %s\n",
			 ptol(pp),ptol(JMPBUF_FP_SSSP (pp->env)),pp->stksize,stkutil(pp),	/*  GeK - JMPBUF macro for portability  */
			 ptol(pp->event),
			 pp->i_state ? 'I' : ' ',
			 (pp->state & WAITING) ? 'W' : ' ',
			 (pp->state & SUSPEND) ? 'S' : ' ',
			 pp->input,pp->output,
			 pp->name) == EOF)
				return 0;
		}
	}
	for(pp = Rdytab;pp != NULLPROC;pp = pp->next){
#ifdef	notdef	/*  GeK - removed env  */
		ep = (struct env *)&pp->env;
#endif	/* notdef */
		if(tprintf("%-10lx%-10lx%-10u%-10u          %c%c%c %2d %2d  %s\n",
		 ptol(pp),ptol(JMPBUF_FP_SSSP (pp->env)),pp->stksize,stkutil(pp),	/*  GeK - JMPBUF macro for portability  */
		 pp->i_state ? 'I' : ' ',
		 (pp->state & WAITING) ? 'W' : ' ',
		 (pp->state & SUSPEND) ? 'S' : ' ',
		 pp->input,pp->output,
		 pp->name) == EOF)
			return 0;
	}
	if(Curproc != NULLPROC){
#ifdef	notdef	/*  GeK - removed env  */
		ep = (struct env *)&Curproc->env;
#endif	/* notdef */
		tprintf("%-10lx%-10lx%-10u%-10u          %c   %2d %2d  %s\n",
		 ptol(Curproc),ptol(JMPBUF_FP_SSSP (Curproc->env)),Curproc->stksize,	/*  GeK - JMPBUF macro for portability  */
		 stkutil(Curproc),
		 Curproc->i_state ? 'I' : ' ',
		 Curproc->input,Curproc->output,
		 Curproc->name);
	}
	return 0;
}
#if	!defined (PROCTRACE) && !defined (PROCLOG)
static
#endif
int
stkutil(pp)
struct proc *pp;
{
	unsigned i;
	register int16 *sp;

	i = pp->stksize;
	/*  GeK - from todo 12/93 - prevent overrunning stack top  */
	for(sp = pp->stack; i != 0 && *sp == STACKPAT; sp++)
		i--;
	return i;
}
/* Return number of used words in interrupt stack. Note hardwired value
 * for stack size; this is also found in the various .asm files
 */
static int
chkintstk()
{
	register int i;
	register int16 *cp;

	for(i=512,cp = Intstk; i != 0 && *cp == STACKPAT; cp++)
		i--;
	return i;
}

/* Verify that stack pointer for current process is within legal limits;
 * also check that no one has dereferenced a null pointer
 */
#ifdef	STACK_CHECK	/*  GeK  */
void
chkstk()
{
	int16 *sbase;
	int16 *stop;
	int16 *sp;

	sp = MK_FP(_SS,_SP);
	if(_SS == _DS){
		/* Probably in interrupt context */
		return;
	}
	sbase = Curproc->stack;
	if(sbase == NULL)
		return;	/* Main task -- too hard to check */

	stop = sbase + Curproc->stksize;
	if(sp < sbase || sp >= stop){
		printf("Stack violation, process %s\n",Curproc->name);
		printf("SP = %lx, legal stack range [%lx,%lx)\n",
		ptol(sp),ptol(sbase),ptol(stop));
		fflush(stdout);
		killself();
	}
	if(*(unsigned short *)NULL != oldNull){
		printf("WARNING: Location 0 smashed, process %s\n",Curproc->name);
		*(unsigned short *)NULL = oldNull;
		fflush(stdout);
	}
}
#endif	/* STACK_CHECK */

/* Machine-dependent initialization of a task */
void
psetup(pp,iarg,parg1,parg2,pc)
struct proc *pp;	/* Pointer to task structure */
int iarg;		/* Generic integer arg */
void *parg1;		/* Generic pointer arg #1 */
void *parg2;		/* Generic pointer arg #2 */
void (*pc) __ARGS((int,void *,void *));		/* Initial execution address */
{
	register int *stktop;
#ifdef	notdef	/*  GeK - removed env  */
	register struct env *ep;
#endif	/* notdef */

	/* Set up stack to make it appear as if the user's function was called
	 * by killself() with the specified arguments. When the user returns,
	 * killself() automatically cleans up.
	 *
	 * First, push args on stack in reverse order, simulating what C
	 * does just before it calls a function.
	 */
	stktop = (int *)(pp->stack + pp->stksize);
#ifdef	LARGEDATA
	*--stktop = FP_SEG(parg2);
#endif
	*--stktop = FP_OFF(parg2);
#ifdef	LARGEDATA
	*--stktop = FP_SEG(parg1);
#endif
	*--stktop = FP_OFF(parg1);
	*--stktop = iarg;
		
	/* Now push the entry address of killself(), simulating the call to
	 * the user function.
	 */
#ifdef	LARGECODE
	*--stktop = FP_SEG(killself);
#endif
	*--stktop = FP_OFF(killself);

	/* Set up task environment. Note that for Turbo-C, the setjmp
	 * sets the interrupt enable flag in the environment so that
	 * interrupts will be enabled when the task runs for the first time.
	 * Note that this requires newproc() to be called with interrupts
	 * enabled!
	 */
	setjmp(pp->env);
#ifdef notdef	/*  GeK - removed env  */
	ep = (struct env *)&pp->env;
	ep->ss = FP_SEG(stktop);
	ep->sp = FP_OFF(stktop);
	ep->cs = FP_SEG(pc);	/* Doesn't hurt in small model */
	ep->ip = FP_OFF(pc);
#endif	/* notdef */
	JMPBUF_SS (pp->env) = FP_SEG(stktop);
	JMPBUF_SP (pp->env) = FP_OFF(stktop);
	JMPBUF_CS (pp->env) = FP_SEG(pc);	/* Doesn't hurt in small model */
	JMPBUF_IP (pp->env) = FP_OFF(pc);
	/*  GeK - from PK's export v951123 (for stktrace.c)  */
	JMPBUF_BP (pp->env) = 0;		/* Anchor stack traces */

	/* Task initially runs with interrupts on */
	pp->i_state = 1;
}
unsigned
phash(event)
void *event;
{
	register unsigned x;

	/* Fold the two halves of the pointer */
	x = FP_SEG(event) ^ FP_OFF(event);

	/* If PHASH is a power of two, this will simply mask off the
	 * higher order bits
	 */
	return x % PHASH;
}
