/*
 *	pop3fngr.c - long finger like listing for Demon's POP3 service
 *
 *	Based in part on sdps.c - hacked from finger.c - GeK	9/1/97
 *
 *	This version uses POP3 and Demon's proprietory *env command
 *
 *	GeK		12/02/97
 *
 *	$Id: pop3fngr.c 1.0 1997/06/23 11:50:36 Graham Exp $
 */

/*
 *  Define this if you have a resolve() that handles multiple A records
 *  and your name's Graham ...
 */
#undef	MULTIHOMED_DNS

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "mbuf.h"
#include "socket.h"
#include "usock.h"
#include "session.h"
#include "proc.h"
#include "netuser.h"
#include "commands.h"
#include "tty.h"
#include "main.h"
#include "cmdparse.h"
#ifdef	MULTIHOMED_DNS
#  include "domain.h"
#endif	/* MULTIHOMED_DNS */

#ifndef	IPPORT_POP3
#  define	IPPORT_POP3	110
#endif	/* ifndef IPPORT_POP3 */

#undef	POP3_TRACING	/*  Define this for console tracing  */
#undef	POP3_LOGGING	/*  Define this for logfile output  */

#define	POP3_CONNECT_TIMEOUT	connect_wait_val	/*  connect timeout  */
#define	POP3_RESPONSE_TIMEOUT	60000UL				/*  1 minute response timeout  */
#define POP3_SERVER	"pop3.demon.co.uk"				/*  hard coded - it ain't gonna work with anyone else's  */
#define BLANK_ADDR	"???"							/*  What to display when no address available  */

#define LINE_BUF_LEN	512		/*  RFC says this ...  */
#define TO_LEN			18		/*  Length of local part of To: envelope to show  */
#define FROM_LEN		45		/*  Ditto but full From: envelope  */

static const char	pop3_server[] = POP3_SERVER;
static const char	blank_addr[] = BLANK_ADDR;

#ifdef	POP3_TRACING
static int	sdps_trace = 0;		/*  Debug Tracing  */
#endif	/* POP3_TRACING */
#ifdef	POP3_LOGGING
static int	sdps_log   = 1;		/*  Debug Logging  */

static FILE	*sdps_logfp = NULL;

static void
log_sdps (int s, char *fmt, ...)
{
	va_list         ap;
	time_t          t;
	struct usock   *up;
	union  sp       local, remote;
	struct tm      *local_tm;

	if (sdps_logfp == NULL)
	{
		sdps_logfp = fopen ("./spool/sdps.log", "at");
		if (sdps_logfp == NULL)
			return;
	}

	t = time (NULL);
#ifdef	notdef
	p = ctime (&t);
	p[24] = '\0';	/*  I hate doing that ...  */
	fprintf (sdps_logfp, "%s ", p);
#endif	/* notdef */
	local_tm = localtime (&t);
	fprintf (sdps_logfp, "%02u/%02u/%02u %02u:%02u:%02u ",
	         local_tm->tm_year % 100,
	         local_tm->tm_mon+1,
	         local_tm->tm_mday,
	         local_tm->tm_hour,
	         local_tm->tm_min,
	         local_tm->tm_sec);

	up = itop (s);
	if (up != NULL)
	{
		local.in  = (struct sockaddr_in *) up->name;
		remote.in = (struct sockaddr_in *) up->peername;
		
		fprintf (sdps_logfp, "%3u %5u %-15s ", s,
		         local.in->sin_port, inet_ntoa (remote.in->sin_addr.s_addr));
	}
	
	va_start (ap, fmt);
	vfprintf (sdps_logfp, fmt, ap);
	va_end (ap);
	fflush (sdps_logfp);
}
#endif	/* POP3_LOGGING */

/*
 *	Gets POP3 response into buffer.
 *
 *	Returns -1 on recv failure
 *	         0 on a + response
 *	         1 on a - response
 *	         2 otherwise
 */
static int
get_pop3_response (int s, char *buffer, int buf_len)
{
	char	*cp;

	alarm (POP3_RESPONSE_TIMEOUT);
	if (recvline (s, buffer, buf_len) > 0)
	{
		alarm (0L);
#ifdef	POP3_TRACING
		if (sdps_trace > 0)
			tprintf ("r %s", buffer);
#endif	/* POP3_TRACING */
#ifdef	POP3_LOGGING
		if (sdps_log > 0)
			log_sdps (s, "r %s", buffer);
#endif	/* POP3_LOGGING */
			
		rip (buffer);

		/*  This should be handled elsewhere  */
		if (buffer[0] == '-')
			tprintf ("Unexpected failure: '%s'\n", buffer);

		return (buffer[0] == '+' ? 0 :
		        buffer[0] == '-' ? 1 :
		                           2);
	}

	cp = !run_timer (&Curproc->alarm) ? "timeout" : sockerr (s);
	alarm (0L);
	tprintf ("Receive failed: %s\n", cp != NULL ? cp : "");
#ifdef	POP3_LOGGING
	if (sdps_log > 0)
		log_sdps (s, "Receive failed: %s\n", cp != NULL ? cp : "");
#endif	/* POP3_LOGGING */

	return (-1);
}

/*
 *	Sends POP3 command
 */
static int
send_pop3_command (int s, char *buffer)
{
#ifdef	POP3_TRACING
	if (sdps_trace > 0)
		tprintf ("s %s", buffer);
#endif	/* POP3_TRACING */
#ifdef	POP3_LOGGING
	if (sdps_log > 0)
		log_sdps (s, "s %s", buffer);
#endif	/* POP3_LOGGING */

	return (usputs (s, buffer));
} 

/*
 *	Reads response to Demon's proprietory *env command
 *
 *	Returns -1 on recv failure
 *	         0 otherwise.
 */
static int
process_env_response (int s, char *buffer, int buf_len, char *to, char *from, int *read)
{	
	int	line_count = 0;

	strcpy (to, blank_addr);
	strcpy (from, blank_addr);
	*read = 0;

	while (recvline (s, buffer, buf_len) > 0)
	{
		line_count++;
		
#ifdef	POP3_TRACING
		if (sdps_trace > 0)
			tprintf ("r %s", buffer);
#endif	/* POP3_TRACING */
#ifdef	POP3_LOGGING
		if (sdps_log > 0)
			log_sdps (s, "r %s", buffer);
#endif	/* POP3_LOGGING */
			
		/*  Dot transparency  */
		if (buffer[0] == '.')
		{
			char	*src, *dst;

			if (buffer[1] == '\n')
				return (0);
				
			src = dst = buffer;
			while ((*(src++) = *(++dst)) != '\0')
				;
		}

		rip (buffer);

		switch (line_count)
		{
			case 1:
				/*  RD or OK  */
				if (buffer[0] == 'r' || buffer[0] == 'R')
					*read = 1;
				break;
			case 3:
				/*  From envelope  */
				strncpy (from, buffer, FROM_LEN);
				from[FROM_LEN] = '\0';
				break;
			case 4:
				{
					/*  To  */
					char	*cp;

					/*  Check and eat any source routeing here ...  */
					cp = strchr (buffer, ':');
					strncpy (to, (cp != NULL) ? cp + 1 : buffer, TO_LEN);
					to[TO_LEN] = '\0';
					break;
				}
			case 2:		/*  line 2 is undefined (nodename)  */
				/*  fall through  */
			default:
				break;
		}
	}
	
	return (-1);
}

int
dosdps (int argc, char *argv[], void *p)
{
	char            to   [TO_LEN+1];
	char            from [FROM_LEN+1];
	struct sockaddr_in sock;
	char           *cp;
	char           *buf;
	int             s = -1;
	struct session *sp;
	int             result;
	int             read;
	int             success = FALSE;
	unsigned long   num_messages;
	unsigned long   total_bytes;
	unsigned long   message_idx;
	unsigned long   size;
#ifdef	MULTIHOMED_DNS
	struct dns_addr_list *addr_list = NULL,	/*  list of server addresses  */
	                     *addr_ptr  = NULL;	/*  list ptr to walk list  */
#endif	/* MULTIHOMED_DNS */

	if (argc < 3)
	{
		/*  Shouldn't happen  */
		return (1);
	}

	/* Allocate a session descriptor */
	if ((sp = newsession (argv[0], TELNET)) == NULL)
	{
		tputs ("Too many sessions\n");
		keywait (NULL, 1);
		return (1);
	}

	buf = mallocw (LINE_BUF_LEN);

	sp->ttystate.echo = sp->ttystate.edit = 0;
	sp->flowmode = 1;
	sock.sin_family = AF_INET;
	sock.sin_port = IPPORT_POP3;

	tprintf ("Resolving %s ... ", pop3_server);

#ifdef	MULTIHOMED_DNS

	/*  Loop until a +OK connection or out of addresses  */
	while (1)
	{
		if (addr_list == NULL)
		{
			addr_list = resolve_list ((char *)pop3_server);
			if (addr_list == NULL)
			{
				tprintf ("Host %s unknown\n", pop3_server);
				goto abort_connection;
			}
			addr_ptr = addr_list;
		}
		else
		{
			addr_ptr = addr_ptr->next;
			if (addr_ptr == NULL)
			{
				tprintf ("Unable to connect to %s\n", pop3_server);
				free (addr_list);
				addr_list = NULL;
				goto abort_connection;
			}
		}

		sock.sin_addr.s_addr = addr_ptr->addr;

		tprintf ("trying %s", psocket ((struct sockaddr *)&sock));
		if ((sp->s = s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
		{
			tprintf("\nCan't create socket\n");
			free (addr_list);
			addr_list = NULL;
			goto abort_connection;
		}
		sockmode (s, SOCK_ASCII);

		alarm (POP3_CONNECT_TIMEOUT);
		if (connect (s, (char *)&sock, sizeof (sock)) < 0)
		{
			if (!run_timer (&Curproc->alarm))
			{
				cp = "timeout";
			}
			else
			{
				alarm (0L);
				cp = sockerr (s);
			}
			tprintf ("\nConnect failed: %s\n", cp != NULL ? cp : "");
			goto next_host;
		}
		alarm (0L);

		tputc ('\n');

		/*  Eat opening banner  */
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result == 0)
			break;	/*  Successful connection  */

next_host:
		close_s (s);
		sp->s = s = -1;

	}	/*  while (1)  */

	free (addr_list);
	addr_list = NULL;
	
	/*  Successful connection here  */

#else	/* !MULTIHOMED_DNS  */

	if ((sock.sin_addr.s_addr = resolve ((char *) pop3_server)) == 0)
	{
		tprintf ("Host %s unknown\n", pop3_server);
		goto abort_connection;
	}
	tprintf ("trying %s", psocket ((struct sockaddr *) &sock));
	if ((sp->s = s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
		tputs("\nCan't create socket\n");
		goto abort_connection;
	}
	sockmode (s, SOCK_ASCII);

	alarm (POP3_CONNECT_TIMEOUT);
	if (connect (s, (char *)&sock, sizeof (sock)) < 0)
	{
		cp = !run_timer (&Curproc->alarm) ? "timeout" : sockerr (s);
		alarm (0L);
		tprintf ("\nConnect failed: %s\n", cp != NULL ? cp : "");
		goto abort_connection;
	}
	alarm (0L);

	tputc ('\n');

	/*  Eat opening banner  */
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

#endif	/*  !MULTIHOMED_DNS  */

	/*  Username  */
	sprintf (buf, "USER %s\n", argv[1]);
	send_pop3_command (s, buf);
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

	/*  Password  */
	sprintf (buf, "PASS %s\n", argv[2]);
	send_pop3_command (s, buf);
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

	tprintf ("Logged in; retrieving mail details for %s ...\n", argv[1]);

#ifdef	notdef
	sdps_trace = 0;
#endif	/* notdef */

	/*  Stats  */
	send_pop3_command (s, "STAT\n");
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;
	result = sscanf (buf, "+OK %lu %lu", &num_messages, &total_bytes);
	if (result != 2)
		goto send_quit;

	tprintf ("%s: %lu messages (%lu bytes)\n\n", argv[1], num_messages, total_bytes);

	for (message_idx = 1; message_idx <= num_messages; message_idx++)
	{
#ifdef	notdef
		/*  UIDL  */
		sprintf (buf, "UIDL %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
#endif	/* notdef */

		/*  Attempt *ENV first as it's the most volatile  */

		/*  *env  */
		sprintf (buf, "*ENV %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
		result = process_env_response (s, buf, LINE_BUF_LEN, to, from, &read);
		if (result != 0)
			goto send_quit;

		/*  Drop any remaining mail domain part in To  */
		/*  NOTE: assumes no source routeing - fixed in process_env_response()  */
		cp = strchr (to, '@');
		if (cp != NULL)
			*cp = '\0';

		/*  Size  */
		sprintf (buf, "LIST %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
		result = sscanf (buf, "+OK %lu %lu", &message_idx, &size);
		if (result != 2)
			goto send_quit;

		tprintf ("%4u %-*s %-*s %7lu %s\n", (unsigned int)(message_idx % 1000), TO_LEN, to, FROM_LEN, from, size, read ? "R" : "");
	}

	tputc ('\n');
	success = TRUE;

send_quit:
	send_pop3_command (s, "QUIT\n");
	if (success)
		get_pop3_response (s, buf, LINE_BUF_LEN);	/*  Wait for response  */

abort_connection:
	if (s > 0)
		close_s (s);
	sp->s = -1;

	free (buf);
	
	keywait (NULL, 1);
	freesession (sp);

	/*  Start SMTP server after successful listing  */
	if (success && argc > 3 && stricmp (argv[3], "smtp") == 0)
	{
		char	buf[16];
		int		save_input, save_output;

		/*
		 *  Substitute input, output sockets so it appears server was
		 *  started from console command session (else smtp output is
		 *  lost when this session terminates).
		 */
		save_input  = Curproc->input;
		save_output = Curproc->output;
		Curproc->input  = Cmdpp->input;
		Curproc->output = Cmdpp->output;

		/*  In a buf as cmdparse mangles the buffer  */
		strcpy (buf, "start smtp\n");
		cmdparse (Cmds, buf, NULL);

		Curproc->input  = save_input;
		Curproc->output = save_output;
	}

#ifdef	POP3_LOGGING
	if (sdps_logfp != NULL)
	{
		fclose (sdps_logfp);
		sdps_logfp = NULL;
	}
#endif	/* POP3_LOGGING */

	return (0);
}
