/* This file contains code to print function/arg stack tracebacks
 * at run time, which is extremely useful for finding heap free() errors.
 *
 * This code is highly specific to Borland C and the 80x6 machines.
 *
 * April 10, 1992 P. Karn
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <time.h>
#include "global.h"
#include "proc.h"
#include "pc.h"		/*  smartdrive_commit_cache()  */

#undef	LINE_NUM
#define FLUSH_CACHE

struct symtab {
	struct symtab *next;
	unsigned short seg;
	unsigned short offs;
	char *name;
};

static struct symtab *Symtab;
static void rdsymtab (int unused,void *name,void *p);
static void clrsymtab (void);
static struct symtab *findsym (void (*)());
static int scompare ();
static void paddr (void (*pc)());

static unsigned short Codeseg;

static FILE *stk_fp = NULL;
#define GEK_notdef

#ifdef	LINE_NUM
struct linenument {
	struct	linenument	*next;
	long				line_num;
	unsigned short		offset;
};

struct linenumtab {
	struct linenumtab	*next;
	char				*segment_name;
	char				*obj_file;
	unsigned short		segment;
	struct linenument	*linenumlist;
};

static struct linenumtab	*Linenumtab;

static struct linenumtab *find_linenumtab (unsigned short text_seg);
static struct linenument *find_linenument (struct linenumtab *lt, unsigned short offset);
static void clrlinenumtab (void);
#endif	/* LINE_NUM */

static int	in_stktrace = 0;

void
stktrace (void)
{
	int i,j;
	unsigned short far *context;
	unsigned short far *cnext;
	unsigned short far *ctmp;
	void (*pc)();
	int nargs;
	struct proc *rdproc;
	struct symtab *sp;
	extern char **_argv;
	char *mapname;
	char *cp;
	time_t t;
#ifdef	LINE_NUM
	struct linenumtab	*lt;
	struct linenument	*le;
#endif	/* LINE_NUM */

	/*  Avoid re-entrancy  */
	if (in_stktrace++ > 0) {
		in_stktrace--;
		return;
	}
	
	if((stk_fp = fopen("stktrace.out","at")) == NULL) {
		in_stktrace--;
		return;	/* Give up */
	}

	time(&t);
	fprintf(stk_fp, "stktrace from proc %s (stksize=%04x stack=%p) at %s",Curproc->name,Curproc->stksize,Curproc->stack,ctime(&t));
	Codeseg = _psp + 0x10;
#ifdef	GEK_notdef
	fprintf(stk_fp, "Code base segment: %04x\n",Codeseg);
#endif
	/* Construct name of map file */
	mapname = malloc(strlen(_argv[0]) + 5);
	strcpy(mapname,_argv[0]);
	if((cp = strrchr(mapname,'.')) != NULL)
		*cp = '\0';
	strcat(mapname,".map");

	/* Read the symbol table in another process to avoid overstressing
	 * the stack in this one
	 */
	rdproc = newproc("rdsymtab",512,rdsymtab,1,mapname,NULL,0);
	pwait(rdproc);
	free(mapname);

	context = MK_FP(_SS,_BP);
	pc = stktrace;

	for(i=0;i<20;i++){
		paddr(pc);
		sp = findsym(pc);
		if(sp != NULL)
			fprintf(stk_fp, " %s+%04x",sp->name,FP_OFF(pc) - sp->offs);

#ifdef	LINE_NUM
		lt = find_linenumtab (FP_SEG(pc));
		if (lt != NULL) {
			fprintf (stk_fp, " %s %s", lt->segment_name, lt->obj_file);
			le = find_linenument (lt, FP_OFF(pc) - (FP_OFF(pc) > sp->offs ? 1 : 0));
			if (le != NULL)
				fprintf (stk_fp, " line %ld", le->line_num);
		}
#endif	/* LINE_NUM */

		if(FP_OFF(context) == 0){
			/* No context left, we're done */
			fputc('\n', stk_fp);
			break;
		}
		cnext = MK_FP(FP_SEG(context),*context);
		/* Compute number of args to display */
		if(FP_OFF(cnext) != 0){
			nargs = (int)(cnext - context) - (1 + (sizeof(pc)/2));
			if(nargs > 20)
				nargs = 20; /* limit to reasonable number */
		} else {
			/* No higher level context, so just print an
			 * arbitrary fixed number of args
			 */
			nargs = 8;
		}
		/* Args start after saved BP and return address */
		ctmp = context + 1 + (sizeof(pc)/2);
		fprintf(stk_fp, " (");
		for(j=0;j<nargs;j++){
			fprintf(stk_fp, "%04x",*ctmp);
			if(j < nargs-1)
				fputc(' ', stk_fp);
			else
				break;
			ctmp++;
		}
		fprintf(stk_fp, ")\n");
#ifdef	notdef
		if(strcmp(cp,"_main") == 0)
			break;
#endif

#ifdef	LARGECODE
		pc = MK_FP(context[2],context[1]);
#else
		pc = (void (*)())MK_FP(FP_SEG(pc),context[1]);
#endif
		context = cnext;
	}
	clrsymtab();
#ifdef	LINE_NUM
	clrlinenumtab();
#endif	/* LINE_NUM */
	fclose(stk_fp);
	stk_fp = NULL;
#ifdef	FLUSH_CACHE
	smartdrive_commit_cache();
#endif
	in_stktrace--;
}

static struct symtab *
findsym (void (*pc)())
{
	struct symtab *sp,*spprev;
	unsigned short seg,offs;
	
#ifdef	LARGECODE
	seg = FP_SEG(pc) - Codeseg;
#else
	seg = 0;	/* Small code, no segment */
#endif
	offs = FP_OFF(pc);
	spprev = NULL;
	for(sp = Symtab;sp != NULL;spprev = sp,sp = sp->next){
		if(sp->seg > seg || (sp->seg == seg && sp->offs > offs)){
			break;
		}
	}
	return spprev;
}

#ifdef	LINE_NUM
static struct linenumtab *
find_linenumtab (unsigned short text_seg)
{
	struct linenumtab *lt;
	unsigned short	seg;
	
	seg = text_seg - Codeseg;
	lt = Linenumtab;
	while (lt != NULL) {
		if (lt->segment == seg)
			return (lt);
		lt = lt->next;
	}
	return (lt);
}

static struct linenument *
find_linenument (struct linenumtab *lt, unsigned short offset)
{
	struct linenument	*le;
	
	if (lt == NULL)
		return (NULL);
	le = lt->linenumlist;
	while (le != NULL && offset < le->offset) {
		le = le->next;
	}
	return (le);
}

static void
clrlinenumtab (void)
{
	struct linenumtab *lt, *ltnext;
	struct linenument *le, *lenext;
	
	for (lt = Linenumtab; lt != NULL; lt = ltnext) {
		ltnext = lt->next;
		
		for (le = lt->linenumlist; le != NULL; le = lenext) {
			lenext = le->next;
			free (le);
		}
		free (lt->segment_name);
		free (lt->obj_file);
		free (lt);
	}
	Linenumtab = NULL;
}
#endif	/* LINE_NUM */

static void
clrsymtab(void)
{
	struct symtab *sp,*spnext;

	for(sp = Symtab;sp != NULL;sp = spnext){
		spnext = sp->next;
		free(sp->name);
		free(sp);
	}
	Symtab = NULL;
}

static void
rdsymtab (int unused, void *name, void *p)
{
	char *buf;
	FILE *fp;
	unsigned short seg;
	unsigned short offs;
	struct symtab *sp;
	struct symtab **spp;
	int size = 0;
	int i;
#ifdef	LINE_NUM
	int	in_line_numbers = 0;
#endif

	if((fp = fopen(name,"rt")) == NULL){
		fprintf(stk_fp, "can't read %s\n",name);
		return;
	}
	buf = (char *)malloc(128);
	if (buf == NULL) {
		fclose (fp);
		return;
	}	
	while(fgets(buf,128,fp),!feof(fp)){
		rip(buf);
		if(strcmp(buf,"  Address         Publics by Value") == 0)
			break;
	}
	if(feof(fp)){
		fprintf(stk_fp, "Can't find header line in %s\n",name);
		free(buf);
		return;
	}
	Symtab = NULL;
	while(fgets(buf,128,fp),!feof(fp)){
		rip(buf);
#ifdef	LINE_NUM
		if (strncmp (buf, "Line numbers for", 16) == 0) {
			char	*p;
			struct linenumtab	*lt;
			
			lt = malloc (sizeof (struct linenumtab));
			lt->segment_name = NULL;
			lt->obj_file = NULL;
			lt->segment = 0;
			lt->linenumlist = NULL;
			lt->next = Linenumtab;
			Linenumtab = lt;

			p = strstr (buf, " segment ");
			*p = '\0';
			p += 9;

			lt->segment_name = strdup (p);
			lt->obj_file = strdup (buf+17);

			in_line_numbers = 1;
			continue;
		}
		else if (in_line_numbers) {
			long	line_num;
			char	*p = buf;
			
			if (Linenumtab == NULL)
				continue;
			
			while (*p != '\0') {
				struct linenument	*le;
				
				if (sscanf (p, "%ld %x:%x", &line_num, &seg, &offs) != 3)
					break;

				if (Linenumtab->segment == 0)
					Linenumtab->segment = seg;
				else if (Linenumtab->segment != seg)
					break;

				le = malloc (sizeof (struct linenument));
				le->line_num = line_num;
				le->offset = offs;
				le->next = Linenumtab->linenumlist;
				Linenumtab->linenumlist = le;

				p += 16;
			}
			
			continue;
		}
#endif	/* LINE_NUM */

		if(sscanf(buf,"%x:%x",&seg,&offs) != 2)
			continue;
		sp = (struct symtab *)malloc(sizeof(struct symtab));
		if (sp == NULL)
			continue;
		sp->offs = offs;
		sp->seg = seg;
		sp->name = strdup(buf+17);
		sp->next = Symtab;
		Symtab = sp;
		size++;
	}
	fclose(fp);
	free(buf);
#ifdef	notdef
	fprintf(stk_fp, "Symbols read: %d\n",size);
#endif
	/* Sort the symbols using the quicksort library function */
	spp = malloc(size*sizeof(struct symtab *));
	for(i=0,sp = Symtab;sp != NULL;i++,sp = sp->next)
		spp[i] = sp;
	qsort(spp,size,sizeof(struct symtab *),scompare);
	/* Now put them back in the linked list */
	Symtab = NULL;
	for(i=size-1;i >= 0;i--){
		sp = spp[i];
		sp->next = Symtab;
		Symtab = sp;
	}
	free(spp);
#ifdef	notdef
	for(sp = Symtab;sp != NULL;sp = sp->next)
		fprintf(stk_fp, "%04x:%04x\t%s\n",sp->seg,sp->offs,sp->name);
#endif
#ifdef	notdef
	{
		struct	linenumtab	*lt;
		struct	linenument	*le;

		lt = Linenumtab;
		while (lt != NULL) {
			fprintf (stk_fp, "Obj %s segment %s %04x\n", lt->obj_file, lt->segment_name, lt->segment);
			i = 0;
			le = lt->linenumlist;
			while (le != NULL) {
				i++;
				if (i == 8) {
					fprintf (stk_fp, "\n");
					i = 0;
				}
				fprintf (stk_fp, " %8ld %04x", le->line_num, le->offset);
				le = le->next;
			}
			fprintf (stk_fp, "\n");
			lt = lt->next;
		}
	}	
#endif
}

static int
scompare (struct symtab **a, struct symtab **b)
{
	if((*a)->seg > (*b)->seg)
		return 1;
	if((*a)->seg < (*b)->seg)
		return -1;
	if((*a)->offs > (*b)->offs)
		return 1;
	if((*a)->offs < (*b)->offs)
		return -1;
	return 0;
}

/* Print a code address according to the memory model */
static void
paddr (void (*pc)())
{
#ifdef	LARGECODE
	fprintf(stk_fp, "%04x:%04x",FP_SEG(pc) - Codeseg,FP_OFF(pc));
#else
	fprintf(stk_fp, "%04x",FP_OFF(pc);
#endif	
}
