/* textwini.c */

/*
   Textwin initialization and termination stuff...
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include "config.h"
#include "global.h"
#include "files.h"
#include "cmdparse.h"
#include "textwin.h"
#include "textwinp.h"



# ifdef TEXTWIN


int bTextwinInitialized = FALSE;
int mouseAvail = 0;


void MouseReset(void)
{
 BEGIN_ASM
        mov ax,00h
        int 33h

        mov mouseAvail, ax
 END_ASM
}


void MouseInit(void)
{
 int mouseYMax = screenHeight * screenCharHeight - 1;
 int mouseXMax = screenWidth  * 8 - 1;

 MouseReset();

 if (mouseAvail)
    {
     BEGIN_ASM
          mov cx,1Fh
          mov dx,offset mouseRoutine
          mov ax,seg mouseRoutine
          mov es,ax
          mov ax,0ch
          int 33h

          mov cx,0
          mov dx, word ptr mouseXMax
          mov ax,07h
          int 33h
    
          mov cx,0
          mov dx, word ptr mouseYMax
          mov ax,08h
          int 33h

          mov cx,0
          mov dx,0
          mov ax,04h
          int 33h

     END_ASM

     ShowMouse();
    }
}


void MouseTerm(void)
{
 MouseReset();
}


typedef struct tagScreenSaver
{
 BYTE      page;
 BYTE      cols;
 BYTE      xpos;
 BYTE      ypos;

 int       mode;          
 int       size;
 PBYTE     screen;
 void FAR *video;

} ScreenSaver;

ScreenSaver originalScreen;


void SaveOriginalScreen(void)
{
 BYTE  xpos;
 BYTE  ypos;
 BYTE  cols,
       page;
 int   mode = 0;


 originalScreen.size  = GetScreenWidth() * GetScreenHeight() * 2;
 originalScreen.video = GetScreenBufferStart();

 if (scrRestore)
    {
     originalScreen.screen = MEMALLOC(originalScreen.size);

     if (originalScreen.screen != NULL)
        _fmemcpy(originalScreen.screen,
                 originalScreen.video,
                 originalScreen.size);
    }
 else
    originalScreen.screen = NULL;

 BEGIN_ASM

   // Get current video mode

   mov ah,0fh
   int 10h
   mov byte ptr mode, al
   mov byte ptr cols, ah
   mov byte ptr page, bh

   // Read cursor position

   mov ah,03h
   mov bh,page
   int 10h
   mov byte ptr xpos, dl
   mov byte ptr ypos, dh

 END_ASM

 if (scrRestore)
    {
     originalScreen.xpos = xpos;
     originalScreen.ypos = ypos;
    }
 else
    {
     originalScreen.xpos = 0;
     originalScreen.ypos = 0;
    }

 originalScreen.mode = mode;
 originalScreen.cols = cols;
 originalScreen.page = page;

 if (forcedToMono)
    {
     // Set the cursor position - directly!

#if 0
     -- this only works if mono is _not_ forced
     BEGIN_ASM

       // Set cursor position

       mov ax,0200h
       mov bx,0000h
       mov dl,byte ptr xpos
       mov dh,byte ptr ypos
       int 10h

     END_ASM
#endif
    }
}



void RestoreOriginalScreen(void)
{
 BYTE xpos = originalScreen.xpos;
 BYTE ypos = originalScreen.ypos;
 BYTE page = originalScreen.page;
 int  mode = originalScreen.mode;

 if (!forcedToMono)
    if (scrMode >= 0) // If the original mode was different, restore it.
       {
        BEGIN_ASM              
                               
          // Set video mode    
                               
          mov ah, 00h          
          mov al, byte ptr mode
          int 10h              
                               
        END_ASM                
       }

 if (originalScreen.screen)
    _fmemcpy(originalScreen.video,
             originalScreen.screen,
             originalScreen.size);
 else
    {
     BEGIN_ASM
       
       // Clear the screen

       mov ah,06h
       mov al,0
       mov bh,07h
       mov cx,0
       mov dl,byte ptr screenWidth
       dec dl
       mov dh,byte ptr screenHeight
       dec dh
       int 10h

     END_ASM
    }

 if (!forcedToMono)
    {
     BEGIN_ASM

       // Select display page

       mov ah,05h
       mov al, byte ptr page
       int 10h

       // Set cursor position

       mov ah, 02h
       mov bh, byte ptr page
       mov dl, byte ptr xpos
       mov dh, byte ptr ypos
       int 10h

     END_ASM
    }
}


void SelectWorkingScreen(void)
{
 BYTE page;
 int  mode = 0;


 if (forcedToMono)
    scrMode = 7;   // Liar!

 if ( (scrMode < 0)                   ||
      (scrMode == originalScreen.mode)  )
    scrMode = -originalScreen.mode;       // Just record the screen mode
 else
 if (!forcedToMono)
    {
     mode = scrMode;
     page = 0;

     BEGIN_ASM

       // Set video mode

       mov ah, 00h
       mov al, byte ptr mode
       int 10h

       // Set video page

       mov ah, 05h
       mov al, byte ptr page
       int 10h

     END_ASM

     GetScreenMetrics(); // As they have just changed...
    }
}


static BOOLEAN InitTextWinModule(void)
{
 BOOLEAN ok = TRUE;
 int cols = GetScreenWidth(),
     rows = GetScreenHeight();


 desktopwindow = PrimAllocWindow(0, 0, cols, rows, cols, rows);

 if (desktopwindow != NULL)
    {
     WINID(desktopwindow) = 1;

     SetWindowAttribute(desktopwindow, defaultWindowAttribute );
     SetEraseCharacter(desktopwindow, param[TW_PARAM_DESKTOP_CHAR]);
     EnableWindowBorder(desktopwindow, FALSE);
     ClearWindow(desktopwindow);
     focuswindow = desktopwindow;
    }
 else
   ok = FALSE;

 return ok;
} /*InitTextWinModule*/


static void TermTextWinModule(void)
{
} /*TermTextWinModule*/


static void InitDisplay(void)
{
 GetScreenMetrics();   
 SaveOriginalScreen(); 
 SelectWorkingScreen();
 MouseInit();          
}


static void TermDisplay(void)
{
 RestoreOriginalScreen();

 if (scrRestore)
    MEMFREE(originalScreen.screen);

 originalScreen.screen = NULL;

 MouseTerm();
}


void TextwinTerm(void)
{
 /*
    This routine is intended to be called automagically at application
    exit time. However, it can be called explicitly whenever you
    feel it is appropriate.
 */

 if (bTextwinInitialized)
    {
     TermTextWinModule();
     bTextwinInitialized = FALSE;

     TermDisplay();
    }
}


void TextwinInit(void)
{
 /*
    If we're not already initialized (this routine might possibly
    be called multiple times) we attempt to install an exit routine.
    If that fails, we don't initialize. If the exit routine installs,
    we initialize textwin.
 */

 if (! bTextwinInitialized)
    if (atexit(TextwinTerm) == 0)
       {                           
        ReadTextwinConfigFile();   
        bTextwinInitialized = TRUE;

        InitDisplay();

        InitTextWinModule();       
       }                           
}


void TextwinSuspend(void)
{
 TermDisplay();
}


void TextwinResume(void)
{
 InitDisplay();

 UpdateWindowsBelowAndIncluding(GetWindowTop());
}


# endif /* TEXTWIN */


/* eof: textwini.c */

