/* textwint.c */

/*
   Textwin's 'terminal' handling routines.
*/

#include "config.h"
#include "global.h"
#include "textwin.h"
#include "textwinp.h"


# ifdef TEXTWIN




static void TTYCarriageReturn(HWINDOW pWindow)
{
 WINCURX(pWindow) = 0;
}

static void ScrollLogical(HWINDOW pWindow, SHORT dx, SHORT dy)
{
 if (IsWindow(pWindow))
    if (dx || dy)
      {
        if (     abs(dx) >= WINLOGICALWIDTH(pWindow)
             ||  abs(dy) >= WINLOGICALHEIGHT(pWindow) )
           ClearWindow(pWindow);
        else
          {
            PUSHORT src;
            PUSHORT dst;
            SHORT   yCtr;

            EnableWindowUpdate(pWindow, FALSE);

# if 0

/*
   Nothing yet uses downward scrolling, so we won't compile it in...
*/

            if (dy > 0)  // Scroll down
              {
                for (yCtr=WINLOGICALHEIGHT(pWindow)-1-dy; yCtr>=0; yCtr--)
                  {
                    src = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr   );
                    dst = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr+dy);

                    memcpy(dst, src, 2*WINLOGICALWIDTH(pWindow));
                  }

                ClearWindowRect( pWindow,
                                 0,
                                 0,
                                 WINLOGICALWIDTH(pWindow),
                                 dy );
              }
            else 

# endif

            if (dy < 0) // Scroll up
              {
                for (yCtr=-dy; yCtr<WINLOGICALHEIGHT(pWindow); yCtr++)
                  {
                    src = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr   );
                    dst = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr+dy);

                    memcpy(dst, src, 2*WINLOGICALWIDTH(pWindow));
                  }

                ClearWindowRect( pWindow,
                                 0,
                                 WINLOGICALHEIGHT(pWindow)+dy,
                                 WINLOGICALWIDTH(pWindow),
                                 -dy );
              }

# if 0
/*
   Nothing yet uses left/right logical buffer scrolling, so we
   won't compile it in...
*/

            if (dx > 0) // Scroll right
              {
                for (yCtr=0; yCtr<WINLOGICALHEIGHT(pWindow); yCtr++)
                  {
                    src = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr);
                    dst = src + dx;

                    memmove(dst, src, 2*(WINLOGICALWIDTH(pWindow)-dx));
                  }

                ClearWindowRect( pWindow,
                                 0,
                                 0,
                                 dx,
                                 WINLOGICALHEIGHT(pWindow) );
              }
            else if (dx < 0) // Scroll left
              {
                for (yCtr=0; yCtr<WINLOGICALHEIGHT(pWindow); yCtr++)
                  {
                    dst = (PUSHORT)&WINCHARXY(pWindow, 0, yCtr);
                    src = dst - dx;

                    memmove(dst, src, 2*(WINLOGICALWIDTH(pWindow)+dx));
                  }

                ClearWindowRect( pWindow,
                                 WINLOGICALWIDTH(pWindow) + dx,
                                 0,
                                 -dx,
                                 WINLOGICALHEIGHT(pWindow) );
              }

# endif

            EnableWindowUpdate(pWindow, TRUE);

            UpdateWindow(pWindow);
          }
      }
} /* ScrollLogical */




static void TTYNewLine(HWINDOW pWindow)
{
 WINCURY(pWindow)++;

 if (WINCURY(pWindow) >= WINLOGICALHEIGHT(pWindow))
    {
     ScrollLogical(pWindow, 0, -1);
     WINCURY(pWindow) = WINLOGICALHEIGHT(pWindow) - 1;
    }

 EnsureCursorIsInWindow(pWindow);
} /*TTYNewLine*/


static void TTYAdvanceCursor(HWINDOW pWindow)
{
 WINCURX(pWindow)++;

 if (WINCURX(pWindow) >= WINLOGICALWIDTH(pWindow))
    {
     TTYCarriageReturn(pWindow);
     TTYNewLine(pWindow);
    }
} /*TTYAdvanceCursor*/


static BOOL TTYWindowBasicPutc(HWINDOW pWindow, char ch);


static void TTYBackspace(HWINDOW pWindow)
{
 WINCURX(pWindow)--;

 if (WINCURX(pWindow) < 0)
    {
     WINCURY(pWindow)--;

     if (WINCURY(pWindow) <= 0)
        {
         WINCURX(pWindow) = 0;
         WINCURY(pWindow) = 0;
        }
     else
        WINCURX(pWindow) = WINLOGICALWIDTH(pWindow) - 1;
    }
} /*TTYBackspace*/


static void TTYDestructiveBackspace(HWINDOW pWindow)
{
 TTYBackspace(pWindow);
 TTYWindowBasicPutc(pWindow, ' ');
 TTYBackspace(pWindow);
} /*TTYDestructiveBackspace*/


static BOOL TTYWindowBasicPutc(HWINDOW pWindow, char ch)
{
 int xpos,
     ypos;
 BOOL bNeedUpdateWindow = FALSE;

 switch (ch)
   {
     case 7:
       BEGIN_ASM
         mov ah,0eh
         mov al,7
         mov bx,0
         int 10h
       END_ASM
       break;

     case 8:
       TTYDestructiveBackspace(pWindow);
       break;

     case '\n':
       TTYNewLine(pWindow);
       break;

     case '\r':
       TTYCarriageReturn(pWindow);
       break;

     case '\t':
       ch = ' ';
       /* DROP THROUGH */

     default:
       xpos = WINCURX(pWindow);
       ypos = WINCURY(pWindow);

       if (   xpos >= 0  &&  xpos < WINLOGICALWIDTH(pWindow)
           && ypos >= 0  &&  ypos < WINLOGICALHEIGHT(pWindow) )
          {
           WINCHARXY(pWindow, xpos, ypos) = (BYTE)ch;
           WINATTRXY(pWindow, xpos, ypos) = (BYTE)WINDEFATTR(pWindow);

           UpdateLogicalRect( pWindow, xpos, xpos, ypos, ypos);
           
           TTYAdvanceCursor(pWindow);
          }

       break;
   }

 return bNeedUpdateWindow;
}


void TTYWindowPutc(HWINDOW pWindow, char ch)
{
 BOOL bNeedUpdateWindow = FALSE;

 if (! pWindow)
   pWindow = desktopwindow;

 if (IsWindow(pWindow))
   {
     bNeedUpdateWindow = TTYWindowBasicPutc(pWindow, ch);

     if (GetWindowFocus() == pWindow)
        UpdateCursorDisplay(pWindow);

     if (bNeedUpdateWindow)
        UpdateWindow(pWindow);
   }
} /*TTYWindowPutc*/


void TTYWindowWrite(HWINDOW pWindow, PSZ s)
{
 BOOL bNeedUpdateWindow = FALSE;

 if (! pWindow)
   pWindow = desktopwindow;

 if (IsWindow(pWindow))
    {
     USHORT ch;

     while ((ch = *s++) != '\0')
           if (TTYWindowBasicPutc(pWindow, ch))
              bNeedUpdateWindow = TRUE;

     if (GetWindowFocus() == pWindow)
        UpdateCursorDisplay(pWindow);

     if (bNeedUpdateWindow)
        UpdateWindow(pWindow);
    }
} /*TTYWindowWrite*/





# endif /* TEXTWIN */


/* eof: textwint.c */

