/* VT102 PC specific routines for VT102 terminal emulator for KA9Q
 *
 * By paul@wolf.demon.co.uk June/July 1992
 */

/*
   Grossly hacked by Anthony McCarthy, 25Jul92, to interface with
   textwin.  If you read this source, please bear in mind that any
   errors introduced are my responsibility. Also, this hack of Paul's
   code is very much an interim version and a properly integrated 
   release is destined to follow in due course...
*/

#include <conio.h>
#include <dos.h>
#include "global.h"
#include "config.h"
#include "session.h"
#include "hardware.h"
#include "pc.h"
#include "vt102.h"
#include "textwin.h"
#include "emul.h"

#define CHARWIDTH(y) (scrn->emulInfo->vt102.line_attrs[(y)]?2:1)
#define WIDTH(x,y) (((x)->width)>>((x)->emulInfo->vt102.line_attrs[y]))

static int	  attrib __ARGS((struct screen *scrn));
static	void video_init __ARGS((struct screen *scrn,int maxlines));
static	void	restore_scrn __ARGS((struct screen *scrn));
static	void	movelines __ARGS((struct screen *scrn,int top,int bottom,int dest));
static	void	movechars __ARGS((struct screen *scrn,int line,int left,int right,int dest));
static int  prim_init_vt102 __ARGS((struct screen * scrn,int maxlines));


extern char far vt102_british[128];
extern char far vt102_usascii[128];
extern char far vt102_specialgraphics[128];


struct	keystr
{
 int val;
 char *str;
};



const struct keystr far vt102keys[] =	 
{ {0x3b00,"\x1bOP"},
  {0x3c00,"\x1bOQ"},
  {0x3d00,"\x1bOR"},
  {0x3e00,"\x1bOS"},
  {0x4700,"\x1b[H"},
  {0x4737,"7"     },
  {0x4800,"\x1b[A"},
  {0x4838,"8"     },
  {0x4939,"9"     },
  {0x4a2d,"-"     },
  {0x4b00,"\x1b[D"},
  {0x4b34,"4"     },
  {0x4c35,"5"     },
  {0x4d00,"\x1b[C"},
  {0x4d36,"6"     },
  {0xe00d,"\r"    },
  {0x4f00,"\x1b[K"},
  {0x4f31,"1"     },
  {0x5000,"\x1b[B"},
  {0x5032,"2"     },
  {0x5133,"3"     },
  {0x5230,"0"     },
  {0x532e,"."     },
  {0x7700,"\x1b[H\x1b[2J"},
  {0x7600,"\x1b[L"},
  {0x8400,"\x1b[M"},
  {0x372A,","     },
  {0}
};

const struct keystr far vt52keys[] =	 
{ {0x3b00,"\x1b?P"},
  {0x3c00,"\x1b?Q"},
  {0x3d00,"\x1b?R"},
  {0x3e00,"\x1b?S"},
  {0x4700,"\x1bH" },
  {0x4800,"\x1b""A" },
  {0x4b00,"\x1b""D" },
  {0x4d00,"\x1b""C" },
  {0x4f00,"\x1bK" },
  {0x5000,"\x1b""B" },
  {0x7500,"\x1bJ" },
  {0x7600,"\x1bH1x\bJ"},
  {0}
};


void video_init(scrn,maxlines)
struct screen *scrn;
int maxlines;
{
	int cols;
	int rows=25;


# ifdef TEXTWIN

	cols    = GetTerminalWidth(scrn->hWnd);
	rows    = GetTerminalHeight(scrn->hWnd);

 scrn->videoRAM = (int far *)GetTerminalBuffer(scrn->hWnd);

# else

	int cols;
	int rows=25;

	_AH=0x0f;
	geninterrupt(0x10);
	cols=_AH;

	if (_AL==7)
		videoRAM=MK_FP(0xb000,0);
	else
	{
		videoRAM=MK_FP(0xb800,0);
		rows=*(char far *)MK_FP(0x40,0x84);
		if ((rows<25)||(rows>60))
			rows=25;
	}

# endif

 if (maxlines > 0)
    rows = min(maxlines, rows); /* VT102 can't cope with big terminals */

	scrn->width =cols;
	scrn->height=rows;
}

void my_textattr(struct screen *scrn, int attr)
{
 scrn->emulInfo->vt102.attrib = attr;

 // SetWindowAttribute(scrn->hWnd, attr);
}

int my_gettextattr(struct screen *scrn)
{
 return scrn->emulInfo->vt102.attrib; // GetWindowAttribute(scrn->hWnd);
}

void	my_gotoxy(scrn,x,y)
struct screen *scrn;
int x;
int y;
{

	scrn->col=x;
	scrn->row=y;

	if (scrn->emulInfo->vt102.line_attrs[y])
	{
		x<<=1;
		if (x>=scrn->width)
		{
			scrn->col=(scrn->width>>1)-1;
			x=scrn->col<<1;
		}
	}

# ifdef TEXTWIN

 TextwinGotoXY(scrn->hWnd, x, y);

# else

	_DH=y;
	_DL=x;
	_BH=0;
	_AH=2;
	geninterrupt(0x10);

# endif

}

void	movelines(scrn,top,bottom,dest)
struct screen * scrn;
int top;
int bottom;
int dest;
{
	int far *destptr=scrn->videoRAM+dest*scrn->width;
	int far *srcptr=scrn->videoRAM+top*scrn->width;
	int len=(bottom-top+1)*scrn->width*2;
	memmove(destptr,srcptr,len);

 AddDirtyYRange(scrn->hWnd, top , bottom-top+1);
 AddDirtyYRange(scrn->hWnd, dest, bottom-top+1);
}

void	movechars(scrn,line,left,right,dest)
struct screen *scrn;
int line;
int left;
int right;
int dest;
{
	int far *ptr1=scrn->videoRAM+line*scrn->width;
	int far *destptr=ptr1+dest*CHARWIDTH(line);
	int far *srcptr=ptr1+left*CHARWIDTH(line);
	int len=(right-left+1)*2*CHARWIDTH(line);
	memmove(destptr,srcptr,len);

 AddDirtyY(scrn->hWnd, line);
}

void    Cursor_Up(scrn,n)
struct screen *scrn;
int n;
{
	int y=my_wherey(scrn)-n;
	if (y<scrn->emulInfo->vt102.scroll_top)
		y=scrn->emulInfo->vt102.scroll_top;
	if (y>scrn->emulInfo->vt102.scroll_bottom)
		y=scrn->emulInfo->vt102.scroll_bottom;

	my_gotoxy(scrn,my_wherex(scrn),y);
}

void	Cursor_Right(scrn,n)
struct screen *scrn;
int n;
{
	int x=my_wherex(scrn)+n;
	if (x<0)
		x=0;
	if (x>=WIDTH(scrn,my_wherey(scrn)))
		x=WIDTH(scrn,my_wherey(scrn))-1;

	my_gotoxy(scrn,x,my_wherey(scrn));
}

void	clr_eol(scrn)
struct screen *scrn;
{
	int y=my_wherey(scrn);
	int x=my_wherex(scrn)*CHARWIDTH(y);

	int *ptr=scrn->videoRAM+(x+y*scrn->width);

	int val=(my_gettextattr(scrn)<<8)+' ';

	while ((x++)<scrn->width)
		*(ptr++)=val;

 AddDirtyCurY(scrn->hWnd);
}

void	clrscrn(scrn)
struct screen *scrn;
{
	int i=scrn->width*scrn->height;
	int *ptr=scrn->videoRAM;
	int val;

	my_gotoxy(scrn,0,0);
	my_textattr(scrn,scrn->emulInfo->vt102.attrib);

	val=(my_gettextattr(scrn)<<8)+' ';

	while (i--)
		*(ptr++)=val;

	if (scrn->emulInfo->vt102.line_attrs!=NULLCHAR)
		  {
     free(scrn->emulInfo->vt102.line_attrs);
		   scrn->emulInfo->vt102.line_attrs = NULLCHAR;
    }

	scrn->emulInfo->vt102.line_attrs=mallocw(scrn->height);

	for (i=0;i<scrn->height;i++)
		scrn->emulInfo->vt102.line_attrs[i]=0;

 AddDirtyRect(scrn->hWnd, 0, 0, scrn->width-1, scrn->height-1);
}

void	clr_eos(scrn)
struct screen *scrn;
{
	int x1=my_wherex(scrn);
	int y1=my_wherey(scrn);
	int y;

	clr_eol(scrn);
	for (y=y1+1;y<scrn->height;y++)
	{
        scrn->emulInfo->vt102.line_attrs[y]=0;
		my_gotoxy(scrn,0,y);
		clr_eol(scrn);
	}
	my_gotoxy(scrn,x1,y1);

 AddDirtyRect(scrn->hWnd, 0, y1, scrn->width-1, scrn->height-1);
}

void	clr_scr(scrn)
struct screen *scrn;
{
	int x=my_wherex(scrn);
	int y=my_wherey(scrn);
	clrscrn(scrn);
	my_gotoxy(scrn,x,y);
}

void	clr_bos(scrn)
struct screen *scrn;
{
	int x1=my_wherex(scrn);
	int y1=my_wherey(scrn);
	int x,
		y;

	for (y=0;y<y1;y++)
	{
        scrn->emulInfo->vt102.line_attrs[y]=0;
		my_gotoxy(scrn,0,y);
		clr_eol(scrn);
	}

	my_gotoxy(scrn,0,y1);
	for (x=0;x<=x1*CHARWIDTH(y1);x++)
		console_putc(scrn,' ');

	my_gotoxy(scrn,x1,y1);
}

void	clr_line(scrn)
struct screen *scrn;
{
	int x1=my_wherex(scrn);
	int y1=my_wherey(scrn);

	my_gotoxy(scrn,0,y1);
	clr_eol(scrn);
	my_gotoxy(scrn,x1,y1);

}

void	clr_bol(scrn)
struct screen *scrn;
{
	int x1=my_wherex(scrn);
	int y1=my_wherey(scrn);
	int x;

	my_gotoxy(scrn,0,y1);
	for (x=0;x<=x1*CHARWIDTH(y1);x++)
		console_putc(scrn,' ');

	my_gotoxy(scrn,x1,y1);
}

void	do_tab(scrn)
struct screen *scrn;
{
	int x1=my_wherex(scrn);
	int y1=my_wherey(scrn);

	do
	{
		console_putc(scrn,' ');
	}
	while ((++x1)%(8*CHARWIDTH(y1)));
}

void	do_lf(scrn)
struct screen *scrn;
{
	int y=my_wherey(scrn)+1;
	int x=my_wherex(scrn);
	int i;

	if (y>scrn->emulInfo->vt102.scroll_bottom)
	{
		movelines(scrn,scrn->emulInfo->vt102.scroll_top+1,scrn->emulInfo->vt102.scroll_bottom,
				  scrn->emulInfo->vt102.scroll_top);

		for (i=scrn->emulInfo->vt102.scroll_top;i<scrn->emulInfo->vt102.scroll_bottom;
				i++)
			scrn->emulInfo->vt102.line_attrs[i]=scrn->emulInfo->vt102.line_attrs[i+1];
		scrn->emulInfo->vt102.line_attrs[i]=0;
		y=scrn->emulInfo->vt102.scroll_bottom;
		my_gotoxy(scrn,0,y);
        clr_eol(scrn);
    }

	my_gotoxy(scrn,x,y);
}

void	rev_index(scrn)
	struct screen *scrn;
{
	int y=my_wherey(scrn)-1;
	int x=my_wherex(scrn);
	int i;

	if (y<scrn->emulInfo->vt102.scroll_top)
	{
		movelines(scrn,scrn->emulInfo->vt102.scroll_top,scrn->emulInfo->vt102.scroll_bottom-1,
				  scrn->emulInfo->vt102.scroll_top+1);

		for (i=scrn->emulInfo->vt102.scroll_bottom;i>scrn->emulInfo->vt102.scroll_top;
				i--)
			scrn->emulInfo->vt102.line_attrs[i]=scrn->emulInfo->vt102.line_attrs[i-1];
		scrn->emulInfo->vt102.line_attrs[i]=0;
		y=scrn->emulInfo->vt102.scroll_top;
		my_gotoxy(scrn,0,y);
		clr_eol(scrn);
    }

	my_gotoxy(scrn,x,y);
}

void    set_xy(scrn,x,y)
struct screen *scrn;
unsigned int x;
unsigned int y;
{
	if (y>scrn->height)
        y=scrn->height;

    if (y<1)
        y=1;

	if (x>WIDTH(scrn,y-1))
		x=WIDTH(scrn,y-1);

	if (x<1)
		x=1;

	my_gotoxy(scrn,x-1,y-1);

}

void	store_posn(scrn)
struct screen *scrn;
{
	scrn->emulInfo->vt102.store_x=my_wherex(scrn);
	scrn->emulInfo->vt102.store_y=my_wherey(scrn);
	scrn->emulInfo->vt102.store_attrib=scrn->emulInfo->vt102.attrib;
}

void	restore_posn(scrn)
struct screen *scrn;
{
	my_gotoxy(scrn,scrn->emulInfo->vt102.store_x,scrn->emulInfo->vt102.store_y);
	my_textattr(scrn,(scrn->emulInfo->vt102.attrib=scrn->emulInfo->vt102.store_attrib));
}


void term_vt102(scrn)
struct screen * scrn;
{
 if (scrn->emulInfo != NULL)
    {
	    if (scrn->emulInfo->vt102.line_attrs != NULLCHAR)
	       	{
          free(scrn->emulInfo->vt102.line_attrs);
          scrn->emulInfo->vt102.line_attrs = NULLCHAR;
         }

     free(scrn->emulInfo);

     scrn->emulInfo = NULL;
    }
}


static int	prim_init_vt102(scrn,maxlines)
struct screen *scrn;
int maxlines;
{
 scrn->emulInfo = (union emulInfo *)mallocw(sizeof(*scrn->emulInfo));

 scrn->emulInfo->vt102.GL = vt102_usascii;
 scrn->emulInfo->vt102.G0 = vt102_usascii;
 scrn->emulInfo->vt102.G1 = vt102_usascii;

	scrn->emulInfo->vt102.line_attrs=NULLCHAR;
#ifndef TEXTWIN
	if (!directvideo)
	{
		scrn->emulInfo->vt102.flags=TTY;
		return 1;
	}
#endif
	video_init(scrn, maxlines);

	scrn->col=0;
	scrn->row=0;

 scrn->emulInfo->vt102.scroll_top    = 0;
	scrn->emulInfo->vt102.scroll_bottom = scrn->height-1;
	scrn->emulInfo->vt102.attrib        = 7;
	scrn->emulInfo->vt102.flags         = ECHO|WRAP|EMUL_ANSI|CURSOR;
	scrn->emulInfo->vt102.bold          = 0;
	scrn->emulInfo->vt102.half          = 0;
	scrn->emulInfo->vt102.blink         = 0;
	scrn->emulInfo->vt102.underline     = 0;
	scrn->emulInfo->vt102.reverse       = 0;
	scrn->emulInfo->vt102.store_x       = 0;
	scrn->emulInfo->vt102.store_y       = 0;
	scrn->emulInfo->vt102.store_attrib  = 0;
	scrn->emulInfo->vt102.print_auto    = 0;

		clrscrn(scrn);
		set_cursor(scrn, 1,0);
		my_textattr(scrn,scrn->emulInfo->vt102.attrib);

# ifdef TEXTWIN
  UpdateWindow(scrn->hWnd);
# endif

 return 1;
}

int	init_vt102(scrn)
struct screen *scrn;
{
 return prim_init_vt102(scrn, 24);
}

int	init_big_vt102(scrn)
struct screen *scrn;
{
 return prim_init_vt102(scrn, 0);
}

void    set_attr(scrn,s)
struct screen *scrn;
char *s;
{
    switch(atoi(s))
    {
        case 0:scrn->emulInfo->vt102.bold      = 0;
               scrn->emulInfo->vt102.half      = 0;
               scrn->emulInfo->vt102.blink     = 0;
               scrn->emulInfo->vt102.underline = 0;
               scrn->emulInfo->vt102.reverse   = 0;
               break;
        case 1:scrn->emulInfo->vt102.bold=1;
			   break;
		case 2:scrn->emulInfo->vt102.half=1;
			   break;
		case 4:scrn->emulInfo->vt102.underline=1;
			   break;
		case 5:scrn->emulInfo->vt102.blink=1;
			   break;
		case 7:scrn->emulInfo->vt102.reverse=1;
			   break;
		case 17:
		case 18:scrn->emulInfo->vt102.bold=0;
				scrn->emulInfo->vt102.half=0;
				break;
		case 20:scrn->emulInfo->vt102.underline=0;
				break;
		case 21:scrn->emulInfo->vt102.blink=0;
				break;
		case 23:scrn->emulInfo->vt102.reverse=0;
				break;
    }
	my_textattr(scrn,(scrn->emulInfo->vt102.attrib=attrib(scrn)));
}

const unsigned char far attr[32]=
						  {0x07,0x0f,0x08,0x0f,0x07,0x0f,0x08,0x0f,
						   0x87,0x8f,0x88,0x8f,0x87,0x8f,0x88,0x8f,
						   0x70,0x7f,0x78,0x7f,0x70,0x7f,0x78,0x7f,
						   0xf0,0xff,0xf8,0xff,0xf0,0xff,0xf8,0xff};
int    attrib(scrn)
struct screen *scrn;
{
	int val=scrn->emulInfo->vt102.bold+(scrn->emulInfo->vt102.half<<1)+(scrn->emulInfo->vt102.underline<<2)+(scrn->emulInfo->vt102.blink<<3)
				+(scrn->emulInfo->vt102.reverse<<4);
	return attr[val];

}

void	set_line_attr(scrn,n)
struct screen *scrn;
int n;
{
	char *ap=&(scrn->emulInfo->vt102.line_attrs[scrn->row]);
	int x=my_wherex(scrn);
	int y=my_wherey(scrn);
	int *ptr;
	unsigned int val;
	int i;

	if (*ap==n)
		return;
	*ap=n;

	ptr=scrn->videoRAM+(y*scrn->width);
	val=(scrn->emulInfo->vt102.attrib<<8)+' ';

	switch(n)
	{
	 case 0:/* Close up double spaced ->single spaced */
			for (i = 0; i < (scrn->width >> 1); i++)
				*(ptr+i)=*(ptr+(i<<1));
			for (;i<scrn->width;i++)
				*(ptr+i)=val;
			break;
	 case 1:/* Open up single spaced -> double spaced */
			for (i=(scrn->width>>1)-1;i>=0;i--)
			{
				*(ptr+(i<<1))=*(ptr+i);
				*(ptr+(i<<1)+1)=val;
			}
            break;
    }
	my_gotoxy(scrn,x,y);

# ifdef TEXTWIN
 AddDirtyY(scrn->hWnd, y);
# endif
}

void	restore_vt102(scrn)
struct screen *scrn;
{
	set_video_mode(scrn);
    restore_scrn(scrn);
    my_gotoxy(scrn,scrn->col,scrn->row);
    my_textattr(scrn,(scrn->emulInfo->vt102.attrib));
	set_cursor(scrn, scrn->emulInfo->vt102.flags&CURSOR,scrn->emulInfo->vt102.flags&VT102_INSERT);
}

void	set_video_mode(scrn)
struct screen *scrn;
{
# if 0
	int mode;
	int oldmode;

	_AH=0xf;
	geninterrupt(0x10);
	mode=oldmode=_AL;
	if (oldmode!=7)
	{
		switch(scrn->width)
		{
		 case 80:mode=3;
				 break;
		 case 132:mode=0x53;
				 break;
		}
		if (oldmode!=mode)
		{
			_AH=0;
			_AL=mode;
			geninterrupt(0x10);
		}
	}
# endif
}

void	restore_scrn(scrn)
struct screen *scrn;
{
# if 0
    if (scrn->save!=NULL)
		memmove(scrn->videoRAM,scrn->save,2*(scrn->height*scrn->width));
    else
        clrscrn(scrn);
# endif
}

void	save_scrn(scrn)
struct screen *scrn;
{
# if 0
	memmove(scrn->save,scrn->videoRAM,2*(scrn->height*scrn->width));
# endif
}

void	ins_char(scrn)
struct screen *scrn;
{
	int x=my_wherex(scrn);
	int y=my_wherey(scrn);

	movechars(scrn,y,x,WIDTH(scrn,y)-2,x+1);
}

void	del_char(scrn,n)
struct screen *scrn;
int n;
{
	int x=my_wherex(scrn);
	int y=my_wherey(scrn);

	movechars(scrn,y,x+n,WIDTH(scrn,y)-1,x);
}

void	ins_line(scrn,n)
	struct screen *scrn;
int n;
{
	int y=my_wherey(scrn);
	int x=my_wherex(scrn);
	int y1=y;

 n = min(scrn->height - scrn->row, n);

 if (n > 0)
    if (y+n < scrn->height)
       {
	       movelines(scrn,y,scrn->height-n-1,y+n);

	       for (y1=scrn->height-1;y1>=(y+n);y1--)
		          scrn->emulInfo->vt102.line_attrs[y1]=scrn->emulInfo->vt102.line_attrs[y1-n];
       }

	y1=y;
	while (n>0)
	{
		my_gotoxy(scrn,0,y1);
		clr_eol(scrn);
		scrn->emulInfo->vt102.line_attrs[y1]=0;
        n--;
		y1++;
	}

	my_gotoxy(scrn,x,y);
}

void	del_line(scrn,n)
	struct screen *scrn;
int n;
{
	int y1;
	int y=my_wherey(scrn);
	int x=my_wherex(scrn);

 n = max(0, min(n, scrn->height - scrn->row));

 if (n > 0)
	   movelines(scrn,y+n,scrn->height-1,y);

	for (y1=y+n;y1<scrn->height;y1++)
		   scrn->emulInfo->vt102.line_attrs[y1-n]=scrn->emulInfo->vt102.line_attrs[y1];

	y1=scrn->height;
	while (n>0)
	{
		my_gotoxy(scrn,0,y1-n);
		clr_eol(scrn);
		scrn->emulInfo->vt102.line_attrs[y1-n]=0;
		n--;
	}
	my_gotoxy(scrn,x,y);
}

void	set_cursor(scrn, n,m)
struct screen *scrn;
int n;
int m;
{
# if 0
	_setcursortype((n)?(m?_SOLIDCURSOR:_NORMALCURSOR):_NOCURSOR);
# endif
}

void	curpos_report(scrn)
struct screen *scrn;
{
	char	temp[10];
	sprintf(temp,"\x1b[%d;%dR",my_wherey(scrn)+1,my_wherex(scrn)+1);
	kbfeed(temp);
}

char *translate_vt102(scrn,c)
struct screen *scrn;
int c;
{
    static char str[4];
	char *cp=NULL;
    int i;
	const struct keystr *keys;

	if ((c==0x5300)||(c==8))
		return "\x7f";

	if (c==0x1c0d)
	{
		if (scrn->emulInfo->vt102.flags&RETURN_CRLF)
			return "\r\n";
		else
			return "\r";
	}

	if (scrn->emulInfo->vt102.flags&EMUL_ANSI)
		keys=vt102keys;
	else
        keys=vt52keys;

    for (i=0;keys[i].val;i++)
    {
        if (c==keys[i].val)
        {
            cp=keys[i].str;
            break;
        }
    }

	if ((cp[1]=='[')&&(scrn->emulInfo->vt102.flags&CURSOR_APPL))
    {
        strcpy(str,cp);
        cp=str;
        str[1]='O';
    }

	if ((scrn->emulInfo->vt102.flags&KEYPAD_APPL)&&(!cp[1]))
	{
		str[0]='\x1b';
		str[1]='O';
		str[2]=*cp;
		str[3]=0;
		cp=str;
	}

	if ((*cp=='\r')&&(scrn->emulInfo->vt102.flags&RETURN_CRLF))
		return "\r\n";

    return cp;
}

#define BELL            '\a'
#define BACKSPACE       '\b'
#define LINEFEED        '\n'

void console_putc(struct screen *scrn,int c)
{
	int col, row, physcol;
	unsigned int videochar;
	c=(unsigned char)c;

	col = my_wherex(scrn);
	row = my_wherey(scrn);

	switch (c)
	{
	 case BELL :
			_AH = 0xe;
			_AL = BELL;
			geninterrupt(0x10);
			break;
	 case BACKSPACE :
			if (col)
				col--;
			break;
	 case CR :
			col = 0;
			break;
	 case LINEFEED :
			do_lf(scrn);
			row=my_wherey(scrn);
			break;
	 default :

   physcol = col * CHARWIDTH(row);

   if (     physcol >= 0  &&  physcol < scrn->width  
        &&  row     >= 0  &&  row     < scrn->height)
      {
			    _AH = my_gettextattr(scrn);
			    _AL = c;
			    videochar = _AX;
       
       AddDirtyXY(scrn->hWnd, col, row);
       
			    *(scrn->videoRAM+(col + row*scrn->width)) = videochar;
      }

   col++;
			break;
	} /* end switch */

	if (col >= WIDTH(scrn,row))
	{
		if (scrn->emulInfo->vt102.flags&WRAP)
		{
			col = 0;
			do_lf(scrn);
			row=my_wherey(scrn);
		}
		else
			col=WIDTH(scrn,row)-1;
    }
	my_gotoxy(scrn,col,row);
}

